open Arglean.Lean
open Arglean.Prematrix
open Arglean.Cop
open Arglean.Trace
open Term
open Mapping
open Utils


module Stats =
struct

(* total number of inferences performed during proof search *)
let infer = ref 0
(* current maximal depth of proof search *)
let depth = ref 0
(* number of inferences performed until last depth change *)
let depthinfer = ref 0
(* currently running strategy *)
let strategy = ref 0

let set_depth d =
  depth := d;
  depthinfer := !infer

let reset_stats () =
  infer := 0;
  depth := 0;
  depthinfer := 0

let print_stats () =
  Format.printf "%% Inf: %i Depth: %i DInf: %i Str: %i\n%!"
    !infer !depth (!infer - !depthinfer) !strategy

end


exception Alarm
exception CleanExit

let setup_signals () =
  ignore (Sys.signal Sys.sigint  (Sys.Signal_handle (fun _ -> raise CleanExit)));
  ignore (Sys.signal Sys.sigterm (Sys.Signal_handle (fun _ -> raise CleanExit)));
  ignore (Sys.signal Sys.sigalrm (Sys.Signal_handle (fun _ -> raise Alarm)))

(* run f a for maximally i seconds (forever if i = 0) *)
let maxsec f = function
    0 -> Lazy.force f
  | i ->
    ignore (Unix.alarm i);
    try let y = Lazy.force f in ignore (Unix.alarm 0); y
    with Alarm -> None

let run_depth f d = Stats.set_depth d; maxsec (lazy (f d)) !depthsec

let iteratively_deepen f max_depth =
  Lazylist.range 0 max_depth
  |> Lazylist.filter_map (run_depth f)
  |> Lazylist.peek

let get_schedule load_db start = function
  0 -> [lazy (iteratively_deepen start 1000)]
| 1 -> (* contiCoP schedule *)
    [ lazy (iteratively_deepen start !maxdepth)
    ; lazy (cut2 := false; cut3 := false; iteratively_deepen start !maxdepth)
    ; lazy (load_db false !defcnf; iteratively_deepen start 1000)
    ]
| 2 -> (* FEMaLeCoP schedule *)
    [ lazy (iteratively_deepen start !maxdepth)
    ; lazy (cut3 := false; maxsec (lazy (iteratively_deepen start !maxdepth)) 10)
    ; lazy (cut2 := false; maxsec (lazy (iteratively_deepen start 1000)) 10)
    ]
| _ -> failwith "unknown schedule"



let run_schedule load_db start =
  Stats.reset_stats ();
  Random.init !Arglean.Random.seed;
  load_db !conj !defcnf;
  Strom.of_list (get_schedule load_db start !schedule) |>
  Strom.mapi (fun i prf ->
    (*if !verbose then Format.printf "Strategy: %i\n%!" (i+1);*)
    Stats.strategy := (i+1); Lazy.force prf) |>
  Strom.filter_map identity |> Strom.next_opt

let print_status = Format.printf "%% SZS status %s\n%!"
let print_error e = print_status ("Error: " ^ e)

let show_result = function
  Some (sub, prf) -> print_status "Theorem";
    if !mlproof then Proof.Clausal.Proof.to_steps ((hash, [])) [] [] prf |> List.rev |> Proof.Clausal.ML.print_proof sub
    else Proof.Clausal.Proof.print_proof_hsh sub ((hash, []), prf)
| None -> print_status "CounterSatisfiable"

let print_error = function
  CleanExit -> print_status "Unknown"
| Failure x -> print_error x
| Parsing.Parse_error -> print_error "Parse_error"
| e -> raise e


module type Search =
sig
  type result
  val process : result option -> unit
  val load_db : Fof.sform -> bool * bool -> unit

  module Make (Subst : Substitution.Substitution) :
    sig
      val start : int -> result option
    end
end

module Cop (S : Search) =
struct

module Searcharray = S.Make (Substitution.Substarray)
module Searchlist  = S.Make (Substitution.Substlist)

module type SSUBST = sig val start : int -> S.result option end

let search file =
  try
    let (module SSubst : SSUBST) =
      if !Arglean.Cop.arraysub
        then (module Searcharray)
        else (module Searchlist) in
    let sform = Sform.sform_of_file file in
    let load_db conj def = S.load_db sform (conj, def) in
    run_schedule load_db SSubst.start |> S.process;
    Stats.print_stats ()
  with e -> print_error e; Stats.print_stats ()

end

let main speclist usage f =
  setup_signals ();

  let tosolve = ref [] in
  Arg.parse speclist (fun s -> tosolve := s :: !tosolve) usage;

  if !tosolve = [] then Arg.usage speclist usage
  else List.iter f (List.rev !tosolve)
