open Cnf;;
open Fof;;

let meson_depth = ref false;;   (* Use depth not inference bound.            *)
let meson_prefine = ref true;;  (* Use Plaisted's positive refinement.       *)
let meson_dcutin = ref 1;;      (* Min size for d-and-c optimization cut-in. *)
let meson_skew = ref 3;;        (* Skew proof bias (one side is <= n / skew) *)

(* ------------------------------------------------------------------------- *)
(* Prolog exception.                                                         *)
(* ------------------------------------------------------------------------- *)

exception Cut;;

(* ------------------------------------------------------------------------- *)
(* Type for recording a MESON proof tree.                                    *)
(* ------------------------------------------------------------------------- *)

type fol_goal =
  Subgoal of (int * term list) * fol_goal list * (int * string) *
             int * (term * int) list;;

let inferences = ref 0;;

(* ----------------------------------------------------------------------- *)
(* Cacheing continuations. Very crude, but it works remarkably well.       *)
(* ----------------------------------------------------------------------- *)

let cacheconts f =
  let memory = ref [] in
  fun (gg,(insts,offset,size) as input) ->
    if List.exists (fun (_,(insts',_,size')) ->
                   insts = insts' & (size <= size' or !meson_depth))
        (!memory)
    then failwith "cachecont"
    else memory := input::(!memory); f input;;

(* ----------------------------------------------------------------------- *)
(* Check ancestor list for repetition.                                     *)
(* ----------------------------------------------------------------------- *)

let checkan insts (p,a) ancestors =
  let p' = -p in
  let t' = (p',a) in
  try let ours = List.assoc p' ancestors in
      if List.exists (fun u -> eq_lit insts t' (snd(fst u))) ours
      then failwith "checkan"
      else ancestors
  with Not_found -> ancestors;;

  (* ----------------------------------------------------------------------- *)
  (* Insert new goal's negation in ancestor clause, given refinement.        *)
  (* ----------------------------------------------------------------------- *)

let rec remove p l =
  match l with
    [] -> failwith "remove"
  | (h::t) -> if p(h) then h,t else
              let y,n = remove p t in y,h::n;;

let insertan insts (p,a) ancestors =
  let p' = -p in
  let t' = (p',a) in
  let ourancp,otheranc =
    try remove (fun (pr,_) -> pr = p') ancestors
    with Failure _ -> (p',[]),ancestors in
  let ouranc = snd ourancp in
  if List.exists (fun u -> eq_lit insts t' (snd(fst u))) ouranc
  then failwith "insertan: loop"
  else (p',(([],t'),(0,"TRUTH"))::ouranc)::otheranc;;

(* ----------------------------------------------------------------------- *)
(* Perform basic MESON expansion.                                          *)
(* ----------------------------------------------------------------------- *)

let meson_single_expand loffset rule ((g,ancestors),(insts,offset,size)) =
  let (hyps,conc),tag = rule in
  let allins = try rev_itlist2 (fol_unify loffset) (snd g) (snd conc) insts with Unify -> failwith "unify" in
  let locin,globin = separate_insts offset insts allins in
  let mk_ihyp h =
    let h' = fol_inst_bump offset locin h in
    h',checkan insts h' ancestors in
  let newhyps = List.map mk_ihyp hyps in
  inferences := !inferences + 1;
  newhyps,(globin,offset+offinc,size-List.length hyps);;

(* ----------------------------------------------------------------------- *)
(* Perform first basic expansion which allows continuation call.           *)
(* ----------------------------------------------------------------------- *)

let rec tryfind f l =
  match l with
      [] -> failwith "tryfind"
    | (h::t) -> try f h with Failure _ -> tryfind f t;;

let meson_expand_cont loffset rules state cont =
  tryfind
   (fun r -> cont (snd r) (meson_single_expand loffset r state)) rules;;

(* ----------------------------------------------------------------------- *)
(* Try expansion and continuation call with ancestor or initial rule.      *)
(* ----------------------------------------------------------------------- *)

let meson_expand rules ((g,ancestors),((insts,offset,size) as tup)) cont =
  let pr = fst g in
  let newancestors = insertan insts g ancestors in
  let newstate = (g,newancestors),tup in
  try if !meson_prefine & pr > 0 then failwith "meson_expand" else
      let arules = List.assoc pr ancestors in
      meson_expand_cont 0 arules newstate cont
  with Cut -> failwith "meson_expand"
     | Not_found | Failure _ ->
      try let crules =
            List.filter (fun ((h,_),_) -> List.length h <= size) (List.assoc pr rules) in
          meson_expand_cont offset crules newstate cont
      with Cut | Not_found | Failure _ -> failwith "meson_expand";;

(* ----------------------------------------------------------------------- *)
(* Simple Prolog engine organizing search and backtracking.                *)
(* ----------------------------------------------------------------------- *)

let rec chop_list n l =
  if n = 0 then [],l else
  try let m,l' = chop_list (n-1) (List.tl l) in (List.hd l)::m,l'
  with Failure _ -> failwith "chop_list";;

let expand_goal rules =
  let rec expand_goal depth ((g,_),(insts,offset,size) as state) cont =
    if depth < 0 then failwith "expand_goal: too deep" else
    meson_expand rules state
      (fun apprule (_,(pinsts,_,_) as newstate) ->
          expand_goals (depth-1) newstate
            (cacheconts(fun (gs,(newinsts,newoffset,newsize)) ->
               let locin,globin = separate_insts offset pinsts newinsts in
               let g' = Subgoal(g,gs,apprule,offset,locin) in
               if globin = insts & gs = [] then
                 try cont(g',(globin,newoffset,size))
                 with Failure _ -> raise Cut
               else
                 try cont(g',(globin,newoffset,newsize))
                 with Cut -> failwith "expand_goal"
                    | Failure _ -> failwith "expand_goal")))

  and expand_goals depth (gl,(insts,offset,size as tup)) cont =
    match gl with
      [] -> cont ([],tup)

    | [g] -> expand_goal depth (g,tup) (fun (g',stup) -> cont([g'],stup))

    | (g::gs as gl) -> if size >= !meson_dcutin then
              let lsize = size / (!meson_skew) in
              let rsize = size - lsize in
              let lgoals,rgoals = chop_list (List.length gl / 2) gl in
              try expand_goals depth (lgoals,(insts,offset,lsize))
                   (cacheconts(fun (lg',(i,off,n)) ->
                       expand_goals depth (rgoals,(i,off,n + rsize))
                         (cacheconts(fun (rg',ztup) -> cont (lg'@rg',ztup)))))
              with Failure _ ->
                  expand_goals depth (rgoals,(insts,offset,lsize))
                    (cacheconts(fun (rg',(i,off,n)) ->
                       expand_goals depth (lgoals,(i,off,n + rsize))
                         (cacheconts (fun (lg',((_,_,fsize) as ztup)) ->
                            if n + rsize <= lsize + fsize
                            then failwith "repetition of demigoal pair"
                            else cont (lg'@rg',ztup)))))
            else
              expand_goal depth (g,tup)
                (cacheconts(fun (g',stup) ->
                    expand_goals depth (gs,stup)
                      (cacheconts(fun (gs',ftup) -> cont(g'::gs',ftup))))) in

  fun g maxdep maxinf cont ->
    expand_goal maxdep (g,([],2 * offinc,maxinf)) cont;;

(* ----------------------------------------------------------------------- *)
(* With iterative deepening of inferences or depth.                        *)
(* ----------------------------------------------------------------------- *)

let solve_goal rules incdepth min max incsize =
  let rec solve n g =
    if n > max then failwith "solve_goal: Too deep";
    (* Format.print_string
        ((string_of_int (!inferences))^" inferences so far. "^
            "Searching with maximum size "^(string_of_int n)^".");
       Format.print_newline()) *)
    try let gi =
          if incdepth then expand_goal rules g n 100000 (fun x -> x)
          else expand_goal rules g 100000 n (fun x -> x) in
         (Format.print_string("solved at "^string_of_int (!inferences));
           Format.print_newline());
        gi
    with Failure _ -> solve (n + incsize) g in
  fun g -> solve min (g,[]);;

(* ----------------------------------------------------------------------- *)
(* Creation of tagged contrapositives from a HOL clause.                   *)
(* This includes any possible support clauses (1 = falsity).               *)
(* The rules are partitioned into association lists.                       *)
(* ----------------------------------------------------------------------- *)

let mk_negated (p,a) = -p,a;;
let rec mk_contraposes n th used unused sofar =
  match unused with
    [] -> sofar
  | h::t -> let nw = (List.map mk_negated (used @ t),h),(n,th) in
            mk_contraposes (n + 1) th (used@[h]) t (nw::sofar);;

let fol_of_mat_line flits =
  let basics = mk_contraposes 0 "a" [] flits [] in
  if List.for_all (fun (p,_) -> p < 0) flits then
    ((List.map mk_negated flits,(1,[])),(-1,"b"))::basics
  else basics;;

let rec itlist f l b =
  match l with
    [] -> b
  | (h::t) -> f h (itlist f t b);;

let rec mem' eq =
  let rec mem x lis =
    match lis with
      [] -> false
    | (h::t) -> eq x h or mem x t
  in mem;;

let insert' eq x l =
  if mem' eq x l then l else x::l;;

let union' eq l1 l2 = itlist (insert' eq) l1 l2;;

let rec uniq l =
  match l with
    x::(y::_ as t) -> let t' = uniq t in
                      if Pervasives.compare x y = 0 then t' else
                      if t'==t then l else x::t'
 | _ -> l;;

let setify s = uniq (List.sort compare s);;

let fol_of_hol_clauses =
  let eqt (a1,(b1,c1)) (a2, (b2,c2)) = false in
(*   ((a1 = a2) & (b1 = b2) & (equals_thm c1 c2)) in*)
  fun thms ->
    let rawrules = itlist (fun x -> union' eqt (fol_of_mat_line x)) thms [] in
    let prs = setify (List.map (fun x -> fst (snd (fst x))) rawrules) in
    let prules =
      List.map (fun t -> t,List.filter (fun x -> ((=) t (fst (snd (fst x))))) rawrules) prs in
    let srules = List.sort (fun (p,_) (q,_) -> compare (abs p) (abs q)) prules in
    srules;;

(* ----------------------------------------------------------------------- *)
(* Basic HOL MESON procedure.                                              *)
(* ----------------------------------------------------------------------- *)

let simple_meson_refute min max inc ths =
  inferences := 0;
  let old_dcutin = !meson_dcutin in
  if !meson_depth then meson_dcutin := 100001 else ();
  let rules = (*optimize_rules*)(fol_of_hol_clauses ths) in
  (try
    let proof,(insts,_,_) = solve_goal rules (!meson_depth) min max inc (1,[]) in
    Printf.printf "SZS status Theorem\n"
   with Failure x -> Printf.printf "\nSZS status Unknown\n%s\n%!" x);
  meson_dcutin := old_dcutin
;;

if Array.length Sys.argv > 1 then
  let ths = file_mat false false true (Sys.argv.(1)) in
  simple_meson_refute 0 50 1 ths;;


