open Term
open Utils

type 'v form =
    Atom of (int * 'v term list)
  | Neg of 'v form
  | Conj of 'v form * 'v form
  | Disj of 'v form * 'v form
  | Impl of 'v form * 'v form
  | Eqiv of 'v form * 'v form
  | Forall of 'v * 'v form
  | Exists of 'v * 'v form

type iform = int form
type sform = string form

let forall x t = Forall (x, t)
let exists x t = Exists (x, t)
let conj l r = Conj (l, r)
let disj l r = Disj (l, r)
let neg t = Neg t
let atom t = Atom t

let form_of_lit ((p, l) as a) =
  if p > 0 then Atom a
  else Neg (Atom (-p, l))

let lit_of_form = function
    Neg (Atom (n, l)) -> -n, l
  | Atom (n, l) -> n, l
  | _ -> failwith "lit_of_form"

let rec map_form_vars f = function
    Atom (p, t) -> Atom (p, List.map (map_term_vars (fun v -> V (f v))) t)
  | Neg t -> Neg (map_form_vars f t)
  | Conj (l, r) -> Conj (map_form_vars f l, map_form_vars f r)
  | Disj (l, r) -> Disj (map_form_vars f l, map_form_vars f r)
  | Eqiv (l, r) -> Eqiv (map_form_vars f l, map_form_vars f r)
  | Impl (l, r) -> Impl (map_form_vars f l, map_form_vars f r)
  | Forall (x, t) -> Forall (f x, map_form_vars f t)
  | Exists (x, t) -> Exists (f x, map_form_vars f t)

let rec fix_form f = function
    x when f x != x -> fix_form f (f x)
  | Forall (x, t) -> Forall (x, fix_form f t)
  | Exists (x, t) -> Exists (x, fix_form f t)
  | Conj (l, r) -> Conj (fix_form f l, fix_form f r)
  | Disj (l, r) -> Disj (fix_form f l, fix_form f r)
  | Eqiv (l, r) -> Eqiv (fix_form f l, fix_form f r)
  | Impl (l, r) -> Impl (fix_form f l, fix_form f r)
  | Neg t -> Neg (fix_form f t)
  | Atom (p, t) -> Atom (p, t)

let unfold_impl = function
    Impl (l, r) -> Disj (Neg l, r)
  | Neg (Impl (l, r)) -> Conj (l, Neg r)
  | x -> x

let unfold_neg = function
    Neg (Neg t) -> t
  | Neg (Forall (i, t)) -> Exists (i, Neg t)
  | Neg (Exists (i, t)) -> Forall (i, Neg t)
  | Neg (Conj (l, r)) -> Disj (Neg l, Neg r)
  | Neg (Disj (l, r)) -> Conj (Neg l, Neg r)
  | x -> x

let unfold_equiv_nano = function
    Eqiv (l, r) -> Conj (Impl (l, r), Impl (r, l))
  | x -> x

let unfold_equiv_lean def = function
    Eqiv (l, r) -> if def then Conj (Impl (l, r), Impl (r, l))
                          else Disj (Conj (l, r), Conj (Neg l, Neg r))
  | Neg (Eqiv (l, r)) -> Disj (Conj (l, Neg r), Conj (Neg l, r))
  | x -> x


(* unbound variables *)
let rec form_vars sf = function
    Forall (i, t) -> if IS.mem i sf then form_vars sf t else IS.remove i (form_vars sf t)
  | Exists (i, t) -> if IS.mem i sf then form_vars sf t else IS.remove i (form_vars sf t)
  | Conj (l, r) -> form_vars (form_vars sf r) l
  | Disj (l, r) -> form_vars (form_vars sf r) l
  | Eqiv (l, r) -> form_vars (form_vars sf r) l
  | Impl (l, r) -> form_vars (form_vars sf r) l
  | Neg t -> form_vars sf t
  | Atom (_, t) -> List.fold_left term_vars sf t

let rec strip_disj sf = function
    Disj (l, r) -> strip_disj (strip_disj sf r) l
  | x -> x :: sf
let rec strip_conj sf = function
    Conj (l, r) -> strip_conj (strip_conj sf r) l
  | x -> x :: sf

let disj_of_forms = function
    h :: t -> List.fold_left (fun acc x -> Disj (acc, x)) h t
  | [] -> failwith "disj_of_forms"

let conj_of_forms = function
    h :: t -> List.fold_left (fun acc x -> Conj (acc, x)) h t
  | [] -> failwith "conj_of_forms"



let rec miniscope = function
    Forall (x, Conj (l, r)) ->
      Conj (miniscope (Forall (x, l)), miniscope (Forall (x, r)))
  | Exists (x, Disj (l, r)) ->
      Disj (miniscope (Exists (x, l)), miniscope (Exists (x, r)))
  | Forall (x, t) -> if IS.mem x (form_vars IS.empty t) then Forall (x, miniscope t) else miniscope t
  | Exists (x, t) -> if IS.mem x (form_vars IS.empty t) then Exists (x, miniscope t) else miniscope t
  | Neg p -> Neg (miniscope p)
  | Conj (l, r) -> Conj (miniscope l, miniscope r)
  | Disj (l, r) -> Disj (miniscope l, miniscope r)
  | x -> x
let rec fix f x = let y = f x in if x = y then x else fix f y
let miniscope = fix miniscope


let rename_form min_conj =
  let rec rename_form ((map, mv) as sf) = function
    Atom (i, l) ->
      let sf, (i, l) = rename_lit sf (i, l) in sf, Atom (i, l)
  | Neg t ->
      let sf, t = rename_form sf t in sf, Neg t
  | Disj (l, r) ->
      let sf, l = rename_form sf l in
      let sf, r = rename_form sf r in sf, Disj (l, r)
  | Conj (l, r) when min_conj ->
      let (_, mvl), l = rename_form sf l in
      let (_, mvr), r = rename_form sf r in (map, max mvl mvr), Conj (l, r)
  | Conj (l, r) when not min_conj ->
      let sf, l = rename_form sf l in
      let sf, r = rename_form sf r in sf, Conj (l, r)
  | Forall (i, t) ->
      let sf, t = rename_form (IM.add i mv map, mv + 1) t in
      sf, Forall (mv, t)
  | Exists (i, t) ->
      let sf, t = rename_form (IM.add i mv map, mv + 1) t in
      sf, Exists (mv, t)
  | _ -> failwith "rename_form"
  in rename_form (IM.empty, 0) %> snd


let nnf t = fix_form unfold_neg t

(*$Q
  Fof_random.arbitrary_form (fun t -> nnf t = nnf (nnf t))
*)

type 'v ex_info =
{ ex_vars   : ('v * 'v ex_info) list
; univ_vars : 'v list
; formula   : 'v form
; position  : string
}

let rec skolem_args ex =
  List.fold_right (List.union2 % skolem_args % snd) ex.ex_vars ex.univ_vars

let rec skolem_progenitors ex =
  List.fold_right (List.union2 % skolem_progenitors % snd) ex.ex_vars [ex]

let ex_info_cmp i1 i2 =
  compare (String.length (i1.position)) (String.length (i2.position))

let skolem_primogenitor = skolem_progenitors %> List.sort ex_info_cmp %> List.hd

type 'v skolem_map = ('v * ('v ex_info * 'v term)) list

let skolem_form const_num skolem_name inner =
  let skolem_tm (ctxt : int skolem_map) =
    map_term_vars (fun i -> try snd (List.assoc i ctxt) with Not_found -> V i) in
  let skolem_of_info ex =
    let primo = skolem_primogenitor ex
    and args = skolem_args ex in
    let relative = String.lchop (String.length primo.position) ex.position in
    A (const_num (skolem_name primo.formula relative), List.map (fun v -> V v) args) in

  let rec skolem (univ : int list) ctxt pos = function
      Atom (p, t) -> Atom (p, List.map (skolem_tm ctxt) t)
    | Neg (Atom _ as a) -> Neg (skolem univ ctxt (pos ^ "0") a)
    | Conj (l, r) -> Conj (skolem univ ctxt (pos ^ "0") l, skolem univ ctxt (pos ^ "1") r)
    | Disj (l, r) -> Disj (skolem univ ctxt (pos ^ "0") l, skolem univ ctxt (pos ^ "1") r)
    | Forall (i, t) -> Forall (i, skolem (i :: univ) ctxt (pos ^ "0") t)
    | Exists (i, t) as ex -> begin
        let free =
          if inner then IS.remove i (form_vars IS.empty t) |> IS.elements
          else univ in
        let ex_vars, univ_vars = List.rev_partition_map
          (fun v -> try Some (v, fst (List.assoc v ctxt)) with Not_found -> None) free in
        let info = {univ_vars = univ_vars; ex_vars = ex_vars; formula = ex; position = pos} in
        skolem univ ((i, (info, skolem_of_info info)) :: ctxt) (pos ^ "0") t
        end
    | _ -> invalid_arg "skolem" in
  skolem [] [] ""


let rec reorder =
  (*let open Arglean.Trace in*)
  let open Big_int in
  function
    (Atom _ as t) | (Neg (Atom _) as t) -> unit_big_int, t
  | Forall (x, t) -> let p, t' = reorder t in p, Forall (x, t')
  | Exists (x, t) -> let p, t' = reorder t in p, Exists (x, t')
  | Disj (l, r) ->
      (*if !verbose then Format.printf "Connector: disj\nLeft: %s\nRight: %s\n%!" (Print.string_of_form l) (Print.string_of_form r);*)
      let (pl, l') = reorder l
      and (pr, r') = reorder r in
      (*if !verbose then Format.printf "(%s, %s) : %s\n%!" (string_of_big_int pl) (string_of_big_int pr) (Print.string_of_form  (Disj (l, r)));*)
      (add_big_int pl pr,
       if gt_big_int pl pr then Disj (r', l') else Disj (l', r'))
  | Conj (l, r) ->
      (*if !verbose then Format.printf "Connector: conj\nLeft: %s\nRight: %s\n%!" (Print.string_of_form l) (Print.string_of_form r);*)
      let (pl, l') = reorder l
      and (pr, r') = reorder r in
      (*if !verbose then Format.printf "(%s, %s) : %s\n%!" (string_of_big_int pl) (string_of_big_int pr) (Print.string_of_form  (Conj (l, r)));*)
      (mult_big_int pl pr,
       if gt_big_int pl pr then Conj (r', l') else Conj (l', r'))
  | _ -> failwith "reorder"


let rec collect_fp ((f, p) as sf) = function
    Neg x -> collect_fp sf x
  | Conj (l, r) -> collect_fp (collect_fp sf l) r
  | Disj (l, r) -> collect_fp (collect_fp sf l) r
  | Eqiv (l, r) -> collect_fp (collect_fp sf l) r
  | Impl (l, r) -> collect_fp (collect_fp sf l) r
  | Forall (i, t) -> collect_fp sf t
  | Exists (i, t) -> collect_fp sf t
  | Atom (i, t) ->
      List.fold_left term_funs f t,
      if t <> [] then IM.add (abs i) (List.length t) p else p

let rec collect_predfunc = function
    Neg x -> collect_predfunc x
  | Conj (l, r) | Disj (l, r) | Eqiv (l, r) | Impl (l, r) ->
      let (pl, fl) = collect_predfunc l
      and (pr, fr) = collect_predfunc r
      in List.union1 pl pr, List.union1 fl fr
  | Forall (i, t) | Exists (i, t) -> collect_predfunc t
  | Atom (_, []) -> ([], [])
  | Atom (i, t) ->
      [abs i, List.length t],
      List.fold_right (fun x -> List.union1 (collect_func x)) t []
