open Lazylist
open Utils

open Arglean.Cop
open Arglean.Trace
open Cnf
open Common
open Database.Clausal.LazyDb
open Mapping
open Term
open Print
open Proof.Clausal.Proof

module Lazysearch : Common.Search =
struct

type result =
  Substitution.Substlist.t * (Term.lit, int) Proof.Clausal.Proof.proof
let process = Option.map Litdata.write_trace %> show_result
let load_db sform (conj, def) =
  mat_of_sform conj def sform |>
  (if !verbose then tap print_matrix else identity) |>
  axioms2db

module Make (Sub : Substitution.Substitution) =
struct

module Subst = Substitution.Substoff (Sub)

(* number of processed literals so far *)
let plits = ref 0

let pred_of_lit (p, _) = if p < 0 then -p, [] else p, []

let cut p xs ys = Lazylist.(if p then cut xs ys else append xs ys)

let rec prove_lit sub (path, lem, lim) lit =
  if !verbosenc then (Format.printf "%d %s\n%!" !plits (string_of_lit (pred_of_lit lit)); incr plits);

  if !verbose then Format.printf "Lit: %s\n%!" (string_of_lit (Subst.inst_lit (fst sub) lit));
  if !verbose then Format.printf "Path: %s\n%!" (string_of_lits (Lazylist.to_list path |> List.map (Subst.inst_lit (fst sub))));
  if !verbose then Format.printf "Lemmas: %s\n%!" (string_of_lits (List.map (Subst.inst_lit (fst sub)) lem));

  let neglit = negate_lit lit in
  let lemmas =
    if !improved_lemmas then begin
      if List.exists (Subst.eq sub lit) lem then
       (if !verbose then Format.printf "lemma\n%!"; cons (sub, lem, Lemma) nil) else nil
    end
    else
      Lazylist.of_list lem |> Lazylist.filter (Subst.eq sub lit) |>
      Lazylist.map (fun l -> if !verbose then Format.printf "lemma\n%!"; (sub, lit :: lem, Lemma))
  and reductions = Lazylist.filter_map
    (fun p ->
      if !verbose then Format.printf "Reduction try %s (for lit %s, lim %d)\n%!" (string_of_lit p) (string_of_lit (Subst.inst_lit (fst sub) lit)) lim;
      Subst.unify sub neglit p >>=
      (fun sub1 -> if !verbose then Format.printf "Reduction works\n%!";
        Some (sub1, lit :: lem, Reduction))
    ) path
  and extensions = db_entries neglit
    |> Lazylist.map (fun ((_, _, vars, hsh) as contra) ->
      if !verbose then Format.printf "Extension try %s (for lit %s, lim %d)\n%!" (Hashtbl.find Database.Clausal.no_contr hsh) (string_of_lit (Subst.inst_lit (fst sub) lit)) lim;
      if lim < 0 && vars > 0 then nil
      else match Subst.unify_rename sub (snd lit) contra with
        Some (sub1, cla1) ->
          if !verbose then Format.printf "Extension clause %s found\n%!" (List.map (Subst.inst_lit (fst sub1)) cla1 |> string_of_clause);
          incr Stats.infer;

          prove_clause sub1 (Lazylist.cons lit path, lem, lim - 1) (cla1, hsh)
          |> Lazylist.map (fun (sub2, prfs) -> (sub2, lit :: lem, Extension (hsh, prfs)))
      | None -> nil)
    |> Lazylist.concat in
  cut !cut1 lemmas (cut !cut2 reductions (cut !cut3 extensions nil))
and prove_clause sub (path, lem, lim) (cl, cl_hsh) = match cl with
    lit :: lits ->
    if (List.exists (fun x -> Lazylist.exists (Subst.eq sub x) path)) cl then (if !verbose then Format.printf "regularity\n%!"; nil)
    else
    prove_lit sub (path, lem, lim) lit
    |> Litdata.trace_proofs lit cl_hsh
    |> Lazylist.map
       (fun (sub1, lem1, prf1) -> prove_clause sub1 (path, lem1, lim) (lits, cl_hsh) |> Lazylist.map
       (fun (sub2, prfs) -> (sub2, (lit, prf1) :: prfs)))
    |> Lazylist.concat
  | [] -> cons (sub, []) nil


let start lim =
  if !verbose || !verbosenc then Format.printf "Start %d\n%!" (lim+1);
  Hashtbl.clear Litdata.cl_out;
  prove_lit (Subst.empty 1000000, 0) (nil, [], lim) (hash, []) |> Lazylist.peek
  >>= (fun (sub, _, prfs) -> Some (fst sub |> Subst.to_list, prfs))

end
end

module Lazycop = Common.Cop (Lazysearch)

let _ =
  let speclist = Arg.align Arglean.(Prematrix.args @ Lean.args @ Cop.args @ Trace.args)
  and usage = "Usage: lazycop [options] <file.p>\nAvailable options are:" in
  Common.main speclist usage Lazycop.search
