open Common
open Fof
open Term
open Ncmatrix
open Strom
open Arglean.Trace
open Arglean.Cop
open Arglean.Nano
open Print
open Ncprint
open Ncproof
open Extclause
open Utils

module Nanosearch : Common.Search =
struct

type jmatrix = int Ncmatrix.matrix
type jklitmat = (int * int) Ncmatrix.litmat
type jkclause = (int * int) Ncmatrix.clause
type prefix = ((int * int list) * int) list
type postfix =
  ((int * int list) * (jmatrix * jmatrix) * int Extclause.litmat_ext)
type result =
  Substitution.Substlist.t *
  (jklitmat * (jklitmat, jkclause, prefix * postfix) Ncproof.proof)

let process = function
  Some (sub, prf) -> print_status "Theorem";
    Ncproof.print_proof_ncl sub prf
| None -> print_status "CounterSatisfiable"

let mat_of_sform conj =
  Prematrix.prematrix_of_sform (unfold_impl %> unfold_equiv_nano) conj !paths %>
  (*tap (List.length %> Format.printf "Number of formulae: %d\n%!") %>
  tap (List.iter (string_of_form %> Format.printf "Formula: %s\n%!")) %>*)
  (*tap (string_of_form %> Format.printf "Formula: %s\n%!") %>*)
  matrix_of_form [] %> index_matrix

let global_mat = ref []

let load_db sform (conj, def) =
  (* always disable conj for now *)
  global_mat := mat_of_sform false sform;
  if !verbose then (
    Format.printf "Matrix:\n%!";
    print_matrix_i !global_mat);
  Ncdatabase.matrix2db !global_mat


module Make (Sub : Substitution.Substitution) =
struct

module Subst = Substitution.Substoff (Sub)

(* number of processed literals so far *)
let plits = ref 0

let unify_vars (sub, off) l1 l2 =
  try Some (Subst.unify_tms sub (List.map var l1) (List.map var l2), off)
  with Subst.Unify -> None

let unify_rename (sub, off) args1 (clext, args2, _, vars) =
  try Some (
    if vars = 0 then ((Subst.unify_tms sub args1 args2, off), clext)
    else
      (Subst.unify_tms_off off sub args1 args2, off + vars),
      offset_clause_ext off clext)
  with Subst.Unify -> None


(* epsilon operator on lists *)
let pick f l =
  let rec go acc = function
    x :: xs -> begin match f x with
      Some y -> (y, fun z -> List.rev_append acc (z::xs))
    | None -> go (x::acc) xs end
  | [] -> failwith "pick: no element found" in
  go [] l

let nth_clause i ((((j, k), v), _) as c) = if i = j then Some c else None
let nth_matrix i = function
    Lit _ -> None
  | Mat ((j, k), _ as m) -> if i = j then Some m else None

let unique f l = List.length (List.filter_map f l) = 1

let rec prove_ec k sub ((i, v), _, lmext as clext) mi pi =
  assert (unique (nth_clause i) mi);
  assert (List.length pi = k);
  if !verbose then Format.printf "prove_ec: ClaB = %s\n%!" (Ncprint.string_of_matrix (fst (matBC_of_clause_ext clext)));
  let ((((i, k1), v1), cla1), miAB) = pick (nth_clause i) mi in
  let alt1 () =
    if !verbosenc then Format.printf "prove_ec: alt1\n%!";
    if !verbose then
      Format.printf "ClaB = %s\n%!" (Ncprint.string_of_matrix (fst (matBC_of_clause_ext clext)));
    if !verbosenc then (match lmext with Matext (_, _, _) -> Format.printf "matext\n%!" | _ -> ());
    match lmext with Matext (j, _, clext') when List.mem (j, k1) pi ->
      if !verbosenc then Format.printf "member(J^K1,PI)\n%!";
      begin match unify_vars sub v v1 with None -> nil | Some sub2 ->
        let (((j, k1'), mi2), claDE) = pick (nth_matrix j) cla1 in
        assert (unique (nth_matrix j) cla1);
        assert (k1 = k1');
        prove_ec k sub2 clext' mi2 pi |> Strom.map
          (fun (sub3, (prefix, postfix), claB1, mi3) ->
            (sub3, (((i, v1), j) :: prefix, postfix), claB1, miAB (((i, k1), v1), claDE (Mat ((j, k1), mi3))))
          )
      end
    | _ -> nil
  and alt2 () =
    if !verbosenc then (
      Format.printf "prove_ec: alt2\n%!";
      if List.mem (i, k1) pi then Format.printf "member\n%!";
      if v = [] then Format.printf "v = []\n%!");
    if List.mem (i, k1) pi && v = [] then nil
    else
      let (claB, claC) = claBC_of_litmat_ext lmext in
      let index cla = (((i, k), v), copy_clause k cla) in
      cons ((sub, ([], clext), index claB, miAB (index claC))) nil in
  Strom.(append (cache alt1) (cache alt2))

let cut p xs ys = Strom.(if p then cut xs ys else append xs ys)

let pred_of_lit (p, _) = if p < 0 then -p, [] else p, []

let rec prove_lit sub (mi, path, pi, lem, lim) lit =
  if !verbosenc then (Format.printf "%d %s\n%!" !plits (string_of_lit (pred_of_lit lit)); incr plits);
  if !verbose then Format.printf "Lit: %s\n%!" (string_of_lit (Subst.inst_lit (fst sub) lit));
  if !verbose then Format.printf "Path: %s\n%!" (string_of_lits (List.map (Subst.inst_lit (fst sub)) path));
  if !verbose then Format.printf "Lemmas: %s\n%!" (string_of_lits (List.map (Subst.inst_lit (fst sub)) lem));
  let neglit = negate_lit lit
  and k = lazy (List.length pi) in
  let lemmas =
    if !improved_lemmas then begin
      if List.exists (Subst.eq sub lit) lem then
       (if !verbose then Format.printf "lemma\n%!"; cons (sub, lem, Lemma) nil) else nil
    end
    else
      Strom.of_list lem |> Strom.filter (Subst.eq sub lit) |>
      Strom.map (fun l -> if !verbose then Format.printf "lemma\n%!"; (sub, lit :: lem, Lemma))
  and reductions = Strom.filter_map
    (fun p -> if !verbose then Format.printf "Reduction try %s\n%!" (string_of_lit (Subst.inst_lit (fst sub) p));
      match Subst.unify sub neglit p with
        Some sub1 -> if !verbose then Format.printf "Reduction works\n%!";
        Some (sub1, lit :: lem, Reduction)
      | None -> None)
    (Strom.of_list path)
  and extensions = Ncdatabase.db_entries neglit
    |> Strom.of_list
    |> Strom.concat_map (fun ((_, args, ground, vars) as contra) ->
      if !verbose then Format.printf "Extension try (for lit %s, args (%s), lim %d, vars %d)\n%!" (string_of_lit (Subst.inst_lit (fst sub) lit)) (String.concat "," (List.map string_of_term args)) lim vars;
      if lim < 0 && not ground then nil
      else match unify_rename sub (snd lit) contra with
        Some (sub1, clext) ->
          if !verbose then Format.printf "Extension works (for lit %s, lim %d)\n%!" (string_of_lit (Subst.inst_lit (fst sub1) lit)) lim;
          if !verbose then Format.printf "Old/new offsets: %d/%d\n%!" (snd sub) (snd sub1);
          incr Stats.infer;
          (*Format.printf "ClaB: %s\n%!" (Ncprint.string_of_matrix (fst (matBC_of_clause_ext clext)));*)
          prove_ec (Lazy.force k) sub1 clext mi pi
          |> Strom.concat_map (fun (sub2, position, ((i, v), claB1), mi1) ->
            if !verbose then Format.printf "Extension clause %s found\n%!" (string_of_nclause claB1);
            prove_clause sub2 (mi1, lit :: path, i::pi, lem, lim - 1) claB1
            |> Strom.map (fun (sub3, prfs) ->
(*
              if verbosenc then Format.printf "Lit__: %s\n%!" (string_of_lit (inst_lit (fst sub3) lit));
              if verbosenc then Format.printf "ClaB1: %s\n%!" (string_of_nclause claB1);
*)
              (sub3, lit :: lem, Extension (Lit neglit :: claB1, position, prfs))))
      | None -> nil) in
  cut (Subst.ground_lit (fst sub) lit)
  (cut !cut1 lemmas (cut !cut2 reductions (cut !cut3 extensions nil))) nil
  |> Strom.map (fun (sub1, lem1, prf1) -> (sub1, lem1, (Lit lit, prf1)))
(* decomposition rule *)
and prove_mat sub (mi, path, pi, lem, lim) (j, mat1) =
  if !verbose then Format.printf "prove_mat\n%!";
  Strom.of_list mat1 |> Strom.concat_map
    (fun ((i, _), cla1) ->
      if !verbose then Format.printf "Decomposition chose: %s\n%!" (string_of_nclause cla1);
      prove_clause sub (mi, path, i::j::pi, lem, lim) cla1
      |> Strom.map (fun (sub1, prf1) -> (sub1, lem, (Mat (j, mat1), Decomposition (cla1, prf1))))
    )
and prove_litmat sub st = map_litmat (prove_lit sub st) (prove_mat sub st)
and prove_clause sub (mi, path, pi, lem, lim) = function
  lm :: cla ->
    let check = List.exists (fun x -> List.exists (Subst.eq sub x) path) in
    let regularity =
      if !improved_regularity then check (clause_lits (lm :: cla))
      else (match lm with Mat _ -> false | Lit l -> check (l :: clause_lits cla)) in
    if regularity then (if !verbose then Format.printf "regularity\n%!"; nil)
    else prove_litmat sub (mi, path, pi, lem, lim) lm |> Strom.concat_map
      (fun (sub1, lem1, prf1) ->
        prove_clause sub1 (mi, path, pi, lem1, lim) cla
        |> Strom.map (fun (sub2, prf2) -> sub2, prf1 :: prf2)
      )
  (* axiom *)
  | [] -> if !verbose then Format.printf "axiom\n%!"; cons (sub, []) nil

let start lim =
  if !verbose || !verbosenc then Format.printf "Start %d\n%!" (lim+1);
  let mati = copy_matrix 0 !global_mat in
  let sub0 = Subst.empty 1000000 in
  Strom.of_list mati |> Strom.filter_map
    (fun (((i, _), v), cla) ->
      clause_positive cla >>= (fun cla1 -> let off = clause_offset cla1 in
      if !verbosenc then Format.printf "Positive clause\n%!";
      if !verbose then Format.printf "New start clause: %s\n%!" (string_of_nclause cla1);
      prove_clause (sub0, off) (mati, [], [i, 0], [], lim) cla1 |>
      Strom.next_opt |>
      Option.map (fun (sub, prfs) -> sub, Decomposition (cla1, prfs)))
    ) |> Strom.next_opt >>= (fun ((sub, off), prf) -> Some (Subst.to_list sub, (Mat ((-1, 0), mati), prf)))

end
end

module Nanocop = Common.Cop (Nanosearch)

let _ =
  let speclist = Arg.align Arglean.(Prematrix.args @ Nano.args @ Cop.args @ Trace.args)
  and usage = "Usage: nanocop [options] <file.p>\nAvailable options are:" in
  Common.main speclist usage Nanocop.search
