open Fof
open Term
open Utils

type 't litmat = Lit of lit | Mat of 't imatrix
and 't imatrix = 't * 't matrix
and 't matrix = 't iclause list
and 't iclause = ('t * int list) * 't clause
and 't clause = 't  litmat list

let map_litmat fl fm = function
    Lit lit -> fl lit
  | Mat mat -> fm mat

let lit x = Lit x
let mat x = Mat x

let lit_of_litmat = function Lit l -> Some l | Mat _ -> None
let clause_lits c = List.filter_map lit_of_litmat c

let clause_of_matrix = function
    [(_, []), c] -> c
  | lm -> [Mat ((), lm)]

let rec matrix_of_form fv = function
    Atom t -> [((), fv), [Lit t]]
  | Neg (Atom t) -> [((), fv), [Lit (negate_lit t)]]
  | Forall (x, t) -> matrix_of_form (x::fv) t
  | Disj (l, r) ->
      let (cl, cr) = Pair.mapn (matrix_of_form [] %> clause_of_matrix) (l, r) in
      [((), fv), cl @ cr]
  | Conj (l, r) ->
      let (ml, mr) = Pair.mapn (matrix_of_form fv) (l, r) in ml @ mr
  | _ -> failwith "matrix_of_form"

let rec form_of_litmat lm = map_litmat form_of_lit form_of_imatrix lm
and form_of_imatrix (_, m) = form_of_matrix m
and form_of_iclause (_, c) = form_of_clause c
and form_of_matrix m = List.map form_of_iclause m |> conj_of_forms
and form_of_clause c = List.map form_of_litmat  c |> disj_of_forms


let rec litmat_max_var acc = map_litmat (lit_max_var acc) (imatrix_max_var acc)
and imatrix_max_var acc (_, m) = matrix_max_var acc m
and iclause_max_var acc (_, c) = clause_max_var acc c
and matrix_max_var acc m = List.fold_left iclause_max_var acc m
and clause_max_var acc c = List.fold_left  litmat_max_var acc c

let  clause_offset c = 1 +  clause_max_var (-1) c
let iclause_offset c = 1 + iclause_max_var (-1) c

let clause_ground c =
  List.for_all (function Lit l -> lit_ground l | Mat _ -> false) c

let rec index_imatrix i (_, m)  = let j, m' = index_matrix (i+1) m in j, (i, m')
and index_matrix i m = List.fold_map index_iclause i m
and index_iclause i ((_, v), c) = let j, c' = index_clause (i+1) c in j, ((i, v), c')
and index_clause i c = List.fold_map index_litmat i c
and index_litmat i = function
    Lit l -> i, Lit l
  | Mat m -> let j, m' = index_imatrix i m in j, Mat m'
let index_matrix m = snd (index_matrix 0 m)

let rec map_imatrix f (t, m) = (f t, map_matrix f m)
and map_matrix f = List.map (map_iclause f)
and map_iclause f ((t, v), c) = ((f t, v), map_clause f c)
and map_clause f = List.map (map_litmat lit (mat % map_imatrix f))

let copy_matrix k = map_matrix (fun i -> i, k)
let copy_clause k = map_clause (fun i -> i, k)


let rec break_map f = function
    [] -> Some []
  | x::xs -> f x >>= fun y -> break_map f xs >>= fun ys -> Some (y::ys)

let rec litmat_positive lm = map_litmat lit_positive imatrix_positive lm
and lit_positive (p, a) = if p < 0 then Some (Lit (p, a)) else None
and imatrix_positive (i, m) =
  match matrix_positive m with [] -> None | m' -> Some (Mat (i, m'))
and matrix_positive m = List.filter_map iclause_positive m
and iclause_positive (i, c) = clause_positive c >>= (fun c' -> Some (i, c'))
and clause_positive c = break_map litmat_positive c

let rec map_litmat_vars f =
  map_litmat (lit % map_lit_vars (var % f)) (mat % map_imatrix_vars f)
and map_imatrix_vars f (i, m) = (i, map_matrix_vars f m)
and map_matrix_vars f m = List.map (map_iclause_vars f) m
and map_iclause_vars f ((i, v), c) = ((i, List.map f v), map_clause_vars f c)
and map_clause_vars f c = List.map (map_litmat_vars f) c

let offset_iclause off = map_iclause_vars ((+) off)
let offset_clause off = map_clause_vars ((+) off)
let offset_matrix off = map_matrix_vars ((+) off)
