%% File: leancop21.pl  -  Version: 2.1  -  Date: 30 Aug 2008
%%
%%         "Make everything as simple as possible, but not simpler."
%%                                                 [Albert Einstein]
%%
%% Purpose: leanCoP: A Lean Connection Prover for Classical Logic
%%
%% Author:  Jens Otten
%% Web:     www.leancop.de
%%
%% Usage: prove(M,P).    % where M is a set of clauses and P is
%%                       %  the returned connection proof
%%                       %  e.g. M=[[q(a)],[-p],[p,-q(X)]]
%%                       %  and  P=[[q(a)],[[-(q(a)),p],[[-(p)]]]]
%%        prove(F,P).    % where F is a first-order formula and
%%                       %  P is the returned connection proof
%%                       %  e.g. F=((p,all X:(p=>q(X)))=>all Y:q(Y))
%%                       %  and  P=[[q(a)],[[-(q(a)),p],[[-(p)]]]]
%%        prove2(F,S,P). % where F is a formula, S is a subset of
%%                       %  [nodef,def,conj,reo(I),scut,cut,comp(J)]
%%                       %  (with numbers I,J) defining attributes
%%                       %  and P is the returned connection proof
%%
%% Copyright: (c) 1999-2008 by Jens Otten
%% License:   GNU General Public License


:- lib(iso).  % load library for ISO compatibility
:- set_flag(occur_check,on).  % global occur check on
:- [def_mm].  % load program for clausal form translation
:- dynamic(pathlim/0), dynamic(lit/3).

% For verbosity, enumerate the literals processed.
:- dynamic(plits/1).

% Number of inferences.
:- dynamic(infs/1).


%%% prove matrix M / formula F

prove(F,Proof) :- prove2(F,[cut,comp(7)],Proof).

prove2(F,Set,Proof) :-
    (F=[_|_] -> M=F ; make_matrix(F,M,Set)),
    retract_all(lit(_,_,_)), (member([-(#)],M) -> S=conj ; S=pos),

    retract_all(plits(_)),
    (member(trace,Set) -> assert(plits(0)) ; true),

    retract_all(infs(_)),
    (member(infs,Set) -> assert(infs(0)) ; true),

    (member(tracep,Set) -> print('Matrix: ['), nl,
      checklist(writeln, M), print(']'), nl ; true),

    assert_clauses(M,S), prove(1,Set,Proof).

prove(PathLim,Set,Proof) :-
    (member(trace,Set) -> print('Start '), print(PathLim), nl ; true),

    \+member(scut,Set) -> prove([-(#)],[],PathLim,[],Set,[Proof]) ;
    lit(#,C,_) -> prove(C,[-(#)],PathLim,[],Set,Proof1),
    Proof=[C|Proof1].
prove(PathLim,Set,Proof) :-
    member(comp(Limit),Set), PathLim=Limit -> prove(1,[],Proof) ;
    (member(comp(_),Set);retract(pathlim)) ->
    PathLim1 is PathLim+1, prove(PathLim1,Set,Proof).

pred_of_lit(L, P) :-
  (L = -X -> L1 = X ; L1 = L),
  L1=..[P|_].


%%% leanCoP core prover

prove([],_,_,_,_,[]).

prove([Lit|Cla],Path,PathLim,Lem,Set,Proof) :-
    Proof=[[[NegLit|Cla1]|Proof1]|Proof2],
    \+ (member(LitC,[Lit|Cla]), member(LitP,Path), LitC==LitP),

    ( member(trace,Set) ->
        (retract(plits(PLits)) ->
          pred_of_lit(Lit, P), print(PLits), print(' '), print(P), nl,
          PLits2 is PLits + 1, assert(plits(PLits2))
        )
        ; true
    ),
    ( member(tracep,Set) ->
        print('Lit': Lit), nl,
        print('Cla': Cla), nl,
        print('Path': Path), nl,
        print('Lem': Lem), nl
      ; true
    ),

    (-NegLit=Lit;-Lit=NegLit) ->
       ( member(LitL,Lem), Lit==LitL, Cla1=[], Proof1=[]
         ,(member(tracep,Set) -> print('Lemma for lit' : Lit), nl ; true)

         ;
         member(NegL,Path), unify_with_occurs_check(NegL,NegLit),
         Cla1=[], Proof1=[]
         ,(member(tracep,Set) -> print('Reduction for lit' : Lit), nl ; true)

         ;
         lit(NegLit,Cla1,Grnd1),
         ( Grnd1=g -> true ; length(Path,K), K<PathLim -> true ;
           \+ pathlim -> assert(pathlim), fail ),

         ( member(infs,Set) ->
             (retract(infs(Infs)) -> Infs2 is Infs + 1, assert(infs(Infs2)))
             ; true
         ),

         ( member(tracep,Set) -> print('Extension clause' : Cla1), nl,
           print('For lit' : Lit), nl ; true),

         prove(Cla1,[Lit|Path],PathLim,Lem,Set,Proof1)
       ),
       ( member(cut,Set) -> ! ; true ),
       prove(Cla,Path,PathLim,[Lit|Lem],Set,Proof2).

%%% write clauses into Prolog's database

assert_clauses([],_).
assert_clauses([C|M],Set) :-
    (Set\=conj, \+member(-_,C) -> C1=[#|C] ; C1=C),
    (ground(C) -> G=g ; G=n), assert_clauses2(C1,[],G),
    assert_clauses(M,Set).

assert_clauses2([],_,_).
assert_clauses2([L|C],C1,G) :-
    append(C1,C,C2), assert(lit(L,C2,G)), append(C1,[L],C3),
    assert_clauses2(C,C3,G).
