%% File: nanocop.pl  -  Version: 1.0  -  Date: 16 January 2016
%%
%% Purpose: nanoCoP: A Non-clausal Connection Prover
%%
%% Author:  Jens Otten
%%
%% Usage:   prove(F,P).  % where F is a first-order formula, e.g.
%%                       %  F=((p,all X:(p=>q(X)))=>all Y:q(Y))
%%                       %  and P is the returned connection proof

:- lib(iso).  % load library for ISO compatibility
:- set_flag(occur_check,on).  % global occur check on

% Inform the interpreter that the definition of the predicate(s)
% may change during execution (using assert/1 and/or retract/1).
:- dynamic(pathlim/0), dynamic(lit/4).

% For verbosity, enumerate the literals processed.
:- dynamic(plits/1).

% Number of inferences.
:- dynamic(infs/1).

% definitions of logical connectives and quantifiers

:- op(1130,xfy,<=>). :- op(1110,xfy,=>). :- op(500, fy,'~').
:- op( 500, fy,all). :- op( 500, fy,ex). :- op(500,xfy,:).

% -----------------------------------------------------------------
% prove(F,Proof) - prove formula F

% prove/2
prove(F,Proof) :- prove2(F,[cut,comp(6)],Proof).

% prove/3
prove2(F,Set,Proof) :-
    bmatrix(F,Set,Mat), retract_all(lit(_,_,_,_)), assert_matrix(Mat),

    retract_all(plits(_)),
    (member(trace,Set) -> assert(plits(0)) ; true),

    retract_all(infs(_)),
    (member(infs,Set) -> assert(infs(0)) ; true),

    (member(tracep,Set) -> print(Mat), nl ; true),

    prove(Mat,1,Set,Proof).

% prove/4
% start rule
prove(Mat,PathLim,Set,[(I^0)^V:Cla1|Proof]) :-
    (member(trace,Set) -> print('Start '), print(PathLim), nl ; true),
    % delete all negative clauses in Cla and store in Cla1
    % Cla1 must be non-empty
    % (! usually is cut, but represents the empty clause from positiveC)
    member((I^0)^V:Cla,Mat), positiveC(Cla,Cla1), Cla1\=!,

    (member(trace,Set) -> print('Positive clause'), nl ; true),
    (member(tracep,Set) -> print('New start clause' : Cla1), nl ; true),

    prove(Cla1,Mat,[],[I^0],PathLim,[],Set,Proof).

% iterative deepening
prove(Mat,PathLim,Set,Proof) :-
    retract(pathlim) ->
    ( member(comp(PathLim),Set) -> prove(Mat,1,[],Proof) ;
      PathLim1 is PathLim+1, prove(Mat,PathLim1,Set,Proof) ) ;
    member(comp(_),Set) -> prove(Mat,1,[],Proof).

pred_of_lit(L, P) :-
  (L = -X -> L1 = X ; L1 = L),
  L1=..[P|_].


% prove/8
% axiom
prove([],_,_,_,_,_,_,[]).

% prove/8
% decomposition rule
prove([J:Mat1|Cla],MI,Path,PI,PathLim,Lem,Set,Proof) :- !,
    % get a clause Cla1 with index I from Mat1 with index/variant J
    member(I^_:Cla1,Mat1),
    % prove Cla1, followed by remaining clause, and append proofs
    % [I,J|PI] is equivalent to I :: J :: PI in ML
    % PI contains indices of clauses and matrices that contain
    % literals on the path
    prove(Cla1,MI,Path,[I,J|PI],PathLim,Lem,Set,Proof1),
    prove(Cla,MI,Path,PI,PathLim,Lem,Set,Proof2),
    append(Proof1,Proof2,Proof).

% reduction and extension rules
prove([Lit|Cla],MI,Path,PI,PathLim,Lem,Set,Proof) :-
    % Lit == LitV iff Lit does not contain variables; used below
    Proof=[[I^V:[NegLit|ClaB1]|Proof1]|Proof2], copy_term(Lit,LitV),

    (member(tracep, Set), member(LitC,[Lit|Cla]), member(LitP,Path), LitC==LitP ->
      print('regularity'), nl ; true),

    % there should not be any shared literals between path and the subgoals (regularity)
    \+ (member(LitC,[Lit|Cla]), member(LitP,Path), LitC==LitP),

    ( member(trace,Set) ->
        (retract(plits(PLits)) ->
          pred_of_lit(Lit, P), print(PLits), print(' '), print(P), nl,
          PLits2 is PLits + 1, assert(plits(PLits2))
        )
        ; true
    ),
    ( member(tracep,Set) ->
        print('Lit': Lit), nl,
        print('Cla': Cla), nl,
        print('Path': Path), nl,
        print('Lem': Lem), nl
      ; true
    ),

    (-NegLit=Lit;-Lit=NegLit) ->
       (member(tracep,Set) -> print('Trying lemma for lit' : Lit), nl ; true),
       ( member(LitL,Lem), Lit==LitL, ClaB1=[], Proof1=[]
         ,(member(tracep,Set) -> print('Lemma for lit' : Lit), nl ; true)
         % lemma rule
         ;
         % reduction rule
         (member(tracep,Set) -> print('Trying reduction for lit' : Lit), nl ; true),
         member(NegL,Path), unify_with_occurs_check(NegL,NegLit),
         ClaB1=[], Proof1=[]
         ,(member(tracep,Set) -> print('Reduction for lit' : Lit), nl ; true)
         ;
         % extension rule
         (member(tracep,Set) -> print('Trying extension for lit' : Lit), nl ; true),
         lit(NegLit,ClaB,Cla1,Grnd1),
         ( Grnd1=g -> true ; length(Path,K), K<PathLim -> true ;
           \+ pathlim -> assert(pathlim), fail ),

         ( member(infs,Set) ->
             (retract(infs(Infs)) -> Infs2 is Infs + 1, assert(infs(Infs2)))
             ; true
         ),

         ( member(tracep,Set) -> print('Extension works for lit' : Lit), nl ; true),

         prove_ec(Set,ClaB,Cla1,MI,PI,I^V:ClaB1,MI1),

         ( member(tracep,Set) -> print('Extension clause' : ClaB1), nl,
           print('For lit' : Lit), nl ; true),

         prove(ClaB1,MI1,[Lit|Path],[I|PI],PathLim,Lem,Set,Proof1)
       ),
       % do not try other options above if cut set or literal contains no vars
       ( (member(cut,Set);Lit==LitV) -> ! ; true ),
       prove(Cla,MI,Path,PI,PathLim,[Lit|Lem],Set,Proof2).

% extension clause (e-clause)
prove_ec(Set,(I^K)^V:ClaB,IV:Cla,MI,PI,ClaB1,MI1) :-
    % demand that matrix MI contains a clause with index I, and
    % that it is preceded by MIA and succeded by MIB
    % furthermore, set copy index to length of path
    (member(tracep,Set) -> print('prove_ec: ClaB' : ((I^K)^V:ClaB)), nl ; true),
    append(MIA,[(I^K1)^V1:Cla1|MIB],MI), length(PI,K),
    ( (member(trace,Set) -> print('prove_ec: alt1'), nl ; true),
      (member(tracep,Set) -> print('ClaB' : ((I^K)^V:ClaB)), nl ; true),
      ClaB=[J^K:[ClaB2]|_],
      Cla=[_:[Cla2|_]|_],
      (member(ecfix,Set) -> ClaB2\==Cla2 ; true),
      (member(trace,Set) -> print('matext'), nl ; true),
      member(J^K1,PI),
      (member(trace,Set) -> print('member(J^K1,PI)'), nl ; true),
      % ClaB starts with a matrix entry that has only a single clause, ClaB2
      unify_with_occurs_check(V,V1),
      % the first clause of the first matrix in Cla is Cla2
      append(ClaD,[J^K1:MI2|ClaE],Cla1),
      prove_ec(Set,ClaB2,Cla2,MI2,PI,ClaB1,MI3),
      append(ClaD,[J^K1:MI3|ClaE],Cla3),
      append(MIA,[(I^K1)^V1:Cla3|MIB],MI1)
      ;
      % it seems that \== does not change V/V1, even if they contain variables
      % that unify -- see e.g. output from PUZ001+2.p
      % (that also seems to be only nonclausal PUZ problem where member(I^K1,PI)
      % becomes true)
        %print('prove_ec with ClaB' : ((I^K)^V: ClaB)), nl,
      (member(trace,Set) ->
        print('prove_ec: alt2'), nl,
        (member(I^K1,PI) -> print('member'), nl ; true),
        (V = [] -> print('v = []'), nl ; true)
        ; true),
      (\+member(I^K1,PI);V\==V1;V\=[]) ->
      ClaB1=(I^K)^V:ClaB, append(MIA,[IV:Cla|MIB],MI1) ).

% -----------------------------------------------------------------
% assert_matrix(Matrix) - write matrix into Prolog's database

% assert_matrix will always fail in the end
% (producing literal entries as side-effect)
assert_matrix(M) :-
    member(IV:C,M), assert_clauses(C,IV:ClaB,ClaB,IV:ClaC,ClaC).
assert_matrix(_).

assert_clauses(C,ClaB,ClaB1,ClaC,ClaC1) :- !,
    append(ClaD,[M|ClaE],C),
    ( M=J:Mat -> append(MatA,[IV:Cla|MatB],Mat),
                 % set ClaB1, and indirectly also ClaB
                 append([J:[IV:ClaB2]|ClaD],ClaE,ClaB1),
                 % make Mat1 with existing MatA/MatB and fresh clause ClaC2
                 append([IV:ClaC2|MatA],MatB,Mat1),
                 % set ClaC1, and indirectly also ClaC
                 append([J:Mat1|ClaD],ClaE,ClaC1),
                 assert_clauses(Cla,ClaB,ClaB2,ClaC,ClaC2)
               % if the matrix M is actually a literal
               ; append(ClaD,ClaE,ClaB1), ClaC1=C,
                 (ground(C) -> Grnd=g ; Grnd=n),
                 %print('Asserting' : (M, ClaB, C, Grnd)), nl,
                 assert(lit(M,ClaB,ClaC,Grnd)), fail ).

% -----------------------------------------------------------------
% bmatrix(Formula,Matrix) - generate indexed matrix

% polarity 0 means positive, polarity 1 means negative (~)
% FV contains the universal quantifier prefix
% FreeV seems to be modified only in beta case
% I, I1, K are used to index and identify copies

bmatrix(F,Set,M) :- univar(F,[],F1), bmatrix(F1,Set,0,_:M,[],[],_,1,_,_).

% only for tracing
bmatrix(F,Set,Pol,M,FreeV,FV,Paths,I,I1,K) :-
  member(tracep,Set),
  %print('Working on' : F), nl,
  F=..[C,F1,F2], (C=',' ; C=(;) ; C=(=>)),
  print('Connector' : C), nl,
  print('Left' : F1), nl,
  print('Right' : F2), nl,
  fail.

bmatrix((F1<=>F2),Set,Pol,M,FreeV,FV,Paths,I,I1,K) :- !,
    bmatrix(((F1=>F2),(F2=>F1)),Set,Pol,M,FreeV,FV,Paths,I,I1,K).

bmatrix((~F),Set,Pol,M,FreeV,FV,Paths,I,I1,K) :- !,
    Pol1 is (1-Pol), bmatrix(F,Set,Pol1,M,FreeV,FV,Paths,I,I1,K).

% universal quantifier
bmatrix(F,Set,Pol,M,FreeV,FV,Paths,I,I1,K) :-
    F=..[C,X:F1], bma(uni,C,Pol), !,
    bmatrix(F1,Set,Pol,M,FreeV,[X|FV],Paths,I,I1,K).

% existential quantifier
bmatrix(F,Set,Pol,M,FreeV,FV,Paths,I,I1,K) :-
    F=..[C,X:F1], bma(exist,C,Pol), !,
    append(FreeV,FV,FreeV1), I2 is I+1,
    copy_term((X,F1,FreeV1),((I^FreeV1),F2,FreeV1)),
    bmatrix(F2,Set,Pol,M,FreeV,FV,Paths,I2,I1,K).

bmatrix(F,Set,Pol,J^K:M3,FreeV,FV,Paths,I,I1,K) :-
    F=..[C,F1,F2], bma(alpha,C,Pol,Pol1,Pol2), !,
    bmatrix(F1,Set,Pol1,J^K:M1,FreeV,FV,Paths1,I,I2,K),
    bmatrix(F2,Set,Pol2,_:M2,FreeV,FV,Paths2,I2,I1,K),
    Paths is Paths1*Paths2,
    (member(tracep,Set) -> print((Paths1, Paths2) : F), nl ; true),
    ((not(member(nopaths,Set)), Paths1>Paths2) ->
      append(M2,M1,M3) ; append(M1,M2,M3)).

bmatrix(F,Set,Pol,I^K:[(I2^K)^FV1:C3],FreeV,FV,Paths,I,I1,K) :-
    F=..[C,F1,F2], bma(beta,C,Pol,Pol1,Pol2), !,
    ( FV=[] -> FV1=FV, F3=F1, F4=F2 ;
      copy_term((FV,F1,F2,FreeV),(FV1,F3,F4,FreeV)) ),
    append(FreeV,FV1,FreeV1),  I2 is I+1, I3 is I+2,
    bmatrix(F3,Set,Pol1,M1,FreeV1,[],Paths1,I3,I4,K),
    bmatrix(F4,Set,Pol2,M2,FreeV1,[],Paths2,I4,I1,K),
    Paths is Paths1+Paths2,
    (member(tracep, Set) -> print((Paths1, Paths2) : F), nl ; true),
    ( (M1=_:[_^[]:C1];[M1]=C1), (M2=_:[_^[]:C2];[M2]=C2) ->
      ( (not(member(nopaths,Set)), Paths1>Paths2) ->
        append(C2,C1,C3) ; append(C1,C2,C3) ) ).

bmatrix(A,Set,0,I^K:[(I2^K)^FV1:[A1]],FreeV,FV,1,I,I1,K)  :-
    copy_term((FV,A,FreeV),(FV1,A1,FreeV)), I2 is I+1, I1 is I+2.
bmatrix(A,Set,1,I^K:[(I2^K)^FV1:[-A1]],FreeV,FV,1,I,I1,K) :-
    copy_term((FV,A,FreeV),(FV1,A1,FreeV)), I2 is I+1, I1 is I+2.

bma(alpha,',',1,1,1). bma(alpha,(;),0,0,0). bma(alpha,(=>),0,1,0).
bma(beta,',',0,0,0).  bma(beta,(;),1,1,1).  bma(beta,(=>),1,0,1).
bma(exist,all,0). bma(exist,ex,1). bma(uni,all,1). bma(uni,ex,0).

% -----------------------------------------------------------------
% positiveC(Clause,ClausePos) - generate positive start clause

positiveC([],[]).
positiveC([M|C],C1) :-
    ( M=I:M1 -> positiveM(M1,M2), M3=I:M2 ; (-_=M->M3=_:[];M3=M) ),
    ( M3=_:[] -> C1=! ; positiveC(C,C2), (C2=!->C1=!;C1=[M3|C2]) ).
positiveM([],[]).
positiveM([I:C|M],M1) :-
    positiveC(C,C1), (C1=! -> M1=M2;M1=[I:C1|M2]), positiveM(M,M2).

% -----------------------------
% create unique variable names

univar(X,_,X)  :- (atomic(X);var(X);X==[[]]), !.
univar(F,Q,F1) :-
    F=..[A,B|T], ( (A=ex;A=all) -> B=X:C, delete2(Q,X,Q1),
    copy_term((X,C,Q1),(Y,D,Q1)), univar(D,[Y|Q],E), F1=..[A,Y:E] ;
    univar(B,Q,B1), univar(T,Q,T1), F1=..[A,B1|T1] ).

% -----------------------------
% delete variable from list

delete2([],_,[]).
delete2([X|T],Y,T1) :- X==Y, !, delete2(T,Y,T1).
delete2([X|T],Y,[X|T1]) :- delete2(T,Y,T1).

