open Term
open Substitution.Substlist
open Utils

module Clausal =
struct

module Step =
struct

type proof_step = Lem of lit | Pat of lit | Res of (lit * lit list * lit list * int)

let is_res_step = function
    Res _ -> true
  | _ -> false

end


module ML =
struct

open Step

let pp_print_res f (lit, path, lem, nlit, npath, nlem, contr) =
  let open Print in
  let print_lits = pp_iter "," pp_print_lit f in
  Format.pp_print_string f (Hashtbl.find Database.Clausal.no_contr contr);
  Format.pp_print_string f ", (";
  pp_print_lit f lit; Format.pp_print_string f "), (";
  print_lits path; Format.pp_print_string f "), (";
  print_lits lem; Format.pp_print_string f "), (";
  pp_print_lit f nlit; Format.pp_print_string f "), (";
  print_lits npath; Format.pp_print_string f "), (";
  print_lits nlem; Format.pp_print_string f ").\n"

let print_res = pp_print_res Format.std_formatter

let res_step sub = function
    Res (lit, path, lem, contr) ->
      let inst = inst_lit sub in
      Some (inst lit, List.map inst path, List.map inst lem, lit, path, lem, contr)
  | _ -> None

let print_proof sub =
  List.filter_map (res_step sub) %> List.rev %> List.iter print_res

end


module Proof =
struct

type ('l, 'c) proof =
    Lemma
  | Reduction
  | Extension of 'c * ('l * ('l, 'c) proof) list

open Print
open Format

let rec pp_print_proof (fl, fc) sub f (lit, prf) =
  pp_open_vbox f 2;
  fl f sub lit; pp_print_char f ' ';
  pp_print_prf (fl, fc) sub f prf;
  pp_close_box f ()
and pp_print_prf (fl, fc) sub f = function
    Reduction -> pp_print_string f "Red"
  | Lemma -> pp_print_string f "Lem"
  | Extension (cl, prfs) ->
      pp_print_string f "Ext "; fc f cl;
      pp_print_proofs (fl, fc) sub f prfs
and pp_print_proofs (fl, fc) sub f =
  List.iter (fun x -> pp_print_cut f (); pp_print_proof (fl, fc) sub f x)

let print_proof (fl, fc) sub prf = pp_nl (pp_print_proof (fl, fc) sub) std_formatter prf

let pp_print_sub_lit f sub = pp_print_lit f % inst_lit sub

let pp_print_contr f =
  Hashtbl.find Database.Clausal.no_rcont %> uncurry List.cons %> Print.pp_print_clause f

let print_proof_hsh prf = print_proof (pp_print_sub_lit, pp_print_contr) prf

let rec of_steps = function
  Step.Lem lit :: tl -> tl, (lit, Lemma)
| Step.Pat lit :: tl -> tl, (lit, Reduction)
| Step.Res (lit, _, _, hsh) :: tl ->
    let (_, cla) = Hashtbl.find Database.Clausal.no_rcont hsh in
    let tl', prfs = List.fold_right_map (fun acc _ -> of_steps acc) tl cla in
    tl', (lit, Extension (hsh, List.rev prfs))
| _ -> failwith "of_steps"

let of_sub_steps (sub, prf) = (sub, prf |> List.rev |> of_steps |> snd |> snd)

let rec to_steps lit pat lem = function
    Lemma -> [Step.Lem lit]
  | Reduction -> [Step.Pat lit]
  | Extension (cl, prfs) ->
      let res = Step.Res (lit, pat, lem, cl) in
      let pat' = lit :: pat in
      let go lem' (l, prf) = l :: lem', to_steps l pat' lem' prf in
      res :: List.concat (List.fold_map go lem prfs |> snd)

end

end
